==> kaffe-0.8.0.tgz: kaffe-0.8.0/kaffe/kaffevm/threadCalls.c <==
/*
 * threadCalls.c
 * Support for threaded ops which may block (read, write, connect, etc.).
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "lerrno.h"
#include "object.h"
#include "thread.h"
#include "md.h"

/*
 * We only need this stuff is we are using the internal thread system.
 */
#if defined(USE_INTERNAL_THREADS)

#define	TH_READ		0
#define	TH_WRITE	1
#define	TH_ACCEPT	TH_READ
#define	TH_CONNECT	TH_WRITE

static int maxFd = -1;
static fd_set readsPending;
static fd_set writesPending;
static thread* readQ[FD_SETSIZE];
static thread* writeQ[FD_SETSIZE];
static struct timeval zerotimeout = { 0, 0 };

int blockOnFile(int, int);
void waitOnEvents(void);

extern thread* currentThread;

/* These are undefined because we do not yet support async I/O */
#undef	F_SETOWN
#undef	FIOSETOWN
#undef	O_ASYNC
#undef	FIOASYNC

/* Select call is indirect so we can change if when embedding Kaffe into
 * something else.
 */
int (*select_call)(int, fd_set*, fd_set*, fd_set*, struct timeval*) = select;

/*
 * Create a threaded file descriptor.
 */
int
threadedFileDescriptor(int fd)
{
#if !defined(BLOCKING_CALLS)
	int r;
	int on = 1;
	int pid;

	/* Make non-blocking */
#if defined(HAVE_FCNTL) && defined(O_NONBLOCK)
	r = fcntl(fd, F_GETFL, 0);
	r = fcntl(fd, F_SETFL, r|O_NONBLOCK);
#elif defined(HAVE_IOCTL) && defined(FIONBIO)
	r = ioctl(fd, FIONBIO, &on);
#else
	r = 0;
#endif
	if (r < 0) {
		return (r);
	}

	/* Allow socket to signal this process when new data is available */
	pid = getpid();
#if defined(HAVE_FCNTL) && defined(F_SETOWN)
	r = fcntl(fd, F_SETOWN, pid);
#elif defined(HAVE_IOCTL) && defined(FIOSETOWN)
	r = ioctl(fd, FIOSETOWN, &pid);
#else
	r = 0;
#endif
	if (r < 0) {
		return (r);
	}

#if defined(HAVE_FCNTL) && defined(O_ASYNC)
	r = fcntl(fd, F_GETFL, 0);
	r = fcntl(fd, F_SETFL, r|O_ASYNC);
#elif defined(HAVE_IOCTL) && defined(FIOASYNC)
	r = ioctl(fd, FIOASYNC, &on);
#else
	r = 0;
#endif
	if (r < 0) {
		return (r);
	}
#endif
	return (fd);
}

/*
 * Threaded create socket.
 */
int
threadedSocket(int af, int type, int proto)
{
	int fd;

	fd = socket(af, type, proto);
	return (threadedFileDescriptor(fd));
}

/*
 * Threaded file open.
 */
int
threadedOpen(const char* path, int flags, int mode)
{
	int fd;

	fd = open(path, flags, mode);
	return (threadedFileDescriptor(fd));
}

/*
 * Threaded socket connect.
 */
int
threadedConnect(int fd, struct sockaddr* addr, size_t len)
{
	int r;

	r = connect(fd, addr, len);
#if !defined(BLOCKING_CALLS)
	if ((r < 0) && (errno == EINPROGRESS || errno == EALREADY || errno == EWOULDBLOCK)) {
		blockOnFile(fd, TH_CONNECT);
		r = 0; /* Assume it's okay when we get released */
	}
#endif

	return (r);
}

/*
 * Threaded socket accept.
 */
int
threadedAccept(int fd, struct sockaddr* addr, size_t* len)
{
	int r;

	for (;;) {
#if defined(BLOCKING_CALLS)
		blockOnFile(fd, TH_ACCEPT);
#endif
		r = accept(fd, addr, len);
		if (r >= 0 || !(errno == EINPROGRESS || errno == EALREADY || errno == EWOULDBLOCK)) {
			break;
		}
#if !defined(BLOCKING_CALLS)
		blockOnFile(fd, TH_ACCEPT);
#endif
	}
	return (threadedFileDescriptor(r));
}

/*
 * Read but only if we can.
 */
ssize_t
threadedRead(int fd, void* buf, size_t len)
{
	ssize_t r;

#if defined(BLOCKING_CALLS)
	if (blockOnFile(fd, TH_READ) < 0) {
		return (-1);
	}
#endif
	for (;;) {
		r = read(fd, buf, len);
		if (r >= 0 || !(errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)) {
			return (r);
		}
		if (blockOnFile(fd, TH_READ) < 0) {
			return (-1);
		}
	}
}

/*
 * Write but only if we can.
 */
ssize_t
threadedWrite(int fd, const void * buf, size_t len)
{
	ssize_t r;
	const void* ptr;

	ptr = buf;
	r = 1;

	while (len > 0 && r > 0) {
#if defined(BLOCKING_CALLS)
		if (blockOnFile(fd, TH_WRITE) < 0) {
			return (-1);
		}
#endif
		r = write(fd, ptr, len);
		if (r >= 0) {
			ptr += r;
			len -= r;
		}
		else if (!(errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)) {
			return (-1);
		}
		else {
#if !defined(BLOCKING_CALLS)
			if (blockOnFile(fd, TH_WRITE) < 0) {
				return (-1);
			}
#endif
			r = 1;
		}
	}
	return (ptr - buf);
}

/*
 * An attempt to access a file would block, so suspend the thread until
 * it will happen.
 */
int
blockOnFile(int fd, int op)
{
	fd_set fset;
	int r;

DBG(	printf("blockOnFile()\n");					)

	/* Trap obviously invalid file descriptors */
	if (fd < 0) {
		errno = EBADF;
		return (-1);
	}

	intsDisable();

	retry:

	/* First a quick check to see if the file handle is usable.
	 * This saves going through all that queuing stuff.
	 */
	FD_ZERO(&fset);
	FD_SET(fd, &fset);
	r = (*select_call)(fd+1, (op == TH_READ ? &fset : 0), (op == TH_WRITE ? &fset : 0), 0, &zerotimeout);
	if (r != 0) {
		intsRestore();
		assert(blockInts == 0);
		return (r);
	}

	if (fd > maxFd) {
		maxFd = fd;
	}
	if (op == TH_READ) {
		FD_SET(fd, &readsPending);
		suspendOnQThread(currentThread, &readQ[fd], NOTIMEOUT);
		FD_CLR(fd, &readsPending);
	}
	else {
		FD_SET(fd, &writesPending);
		suspendOnQThread(currentThread, &writeQ[fd], NOTIMEOUT);
		FD_CLR(fd, &writesPending);
	}

	/* If we have an error flagged, retry the whole thing. */
	if ((TCTX(currentThread)->flags & THREAD_FLAGS_ERROR) != 0) {
		TCTX(currentThread)->flags &= ~THREAD_FLAGS_ERROR;
		goto retry;
	}

	intsRestore();
	assert(blockInts == 0);

	return (1);
}

/*
 * Check if some file descriptor or other event to become ready.
 * Block if required (but make sure we can still take timer interrupts).
 */
void
checkEvents(bool block)
{
	int r;
	fd_set rd;
	fd_set wr;
	thread* tid;
	thread* ntid;
	int i;
	int b;

DBG(	printf("checkEvents block:%d\n", block);			)

	FD_COPY(&readsPending, &rd);
	FD_COPY(&writesPending, &wr);

	/*
	 * If select() is called with indefinite wait, we have to make sure
	 * we can get interrupted by timer events.
	 */
	if (block == true) {
		b = blockInts;
		blockInts = 0;
		r = (*select_call)(maxFd+1, &rd, &wr, 0, 0);
		blockInts = b;
	}
	else {
		r = (*select_call)(maxFd+1, &rd, &wr, 0, &zerotimeout);
	}

	/* We must be holding off interrupts before we start playing with
	 * the read and write queues.  This should be already done but a
	 * quick check never hurt anyone.
	 */
	assert(blockInts > 0);

	/* On a select error, mark all threads in error and release
	 * them all.  They should try to re-select and pick up their
	 * own individual errors.
	 */
	if (r < 0 && errno != EINTR) {
		for (i = 0; r > 0 && i <= maxFd; i++) {
			for (tid = readQ[i]; tid != 0; tid = ntid) {
				ntid = tid->next;
				TCTX(tid)->flags |= THREAD_FLAGS_ERROR;
				iresumeThread(tid);
			}
			for (tid = writeQ[i]; tid != 0; tid = ntid) {
				ntid = tid->next;
				TCTX(tid)->flags |= THREAD_FLAGS_ERROR;
				iresumeThread(tid);
			}
			writeQ[i] = 0;
			readQ[i] = 0;
		}
		return;
	}

DBG(	printf("Select returns %d\n", r);				)

	for (i = 0; r > 0 && i <= maxFd; i++) {
		if (readQ[i] != 0 && FD_ISSET(i, &rd)) {
			for (tid = readQ[i]; tid != 0; tid = ntid) {
				ntid = tid->next;
				iresumeThread(tid);
			}
			readQ[i] = 0;
			r--;
		}
		if (writeQ[i] != 0 && FD_ISSET(i, &wr)) {
			for (tid = writeQ[i]; tid != 0; tid = ntid) {
				ntid = tid->next;
				iresumeThread(tid);
			}
			writeQ[i] = 0;
			r--;
		}
	}
}
#endif

==> kaffe-0.8.1.tgz: kaffe-0.8.1/kaffe/kaffevm/threadCalls.c <==
/*
 * threadCalls.c
 * Support for threaded ops which may block (read, write, connect, etc.).
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>, 1996.
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "lerrno.h"
#include "object.h"
#include "thread.h"
#include "md.h"

/*
 * We only need this stuff is we are using the internal thread system.
 */
#if defined(USE_INTERNAL_THREADS)

/* These are undefined because we do not yet support async I/O */
#undef	F_SETOWN
#undef	FIOSETOWN
#undef	O_ASYNC
#undef	FIOASYNC

/*
 * Create a threaded file descriptor.
 */
int
threadedFileDescriptor(int fd)
{
#if !defined(BLOCKING_CALLS)
	int r;
	int on = 1;
	int pid;

	/* Make non-blocking */
#if defined(HAVE_FCNTL) && defined(O_NONBLOCK)
	r = fcntl(fd, F_GETFL, 0);
	r = fcntl(fd, F_SETFL, r|O_NONBLOCK);
#elif defined(HAVE_IOCTL) && defined(FIONBIO)
	r = ioctl(fd, FIONBIO, &on);
#else
	r = 0;
#endif
	if (r < 0) {
		return (r);
	}

	/* Allow socket to signal this process when new data is available */
	pid = getpid();
#if defined(HAVE_FCNTL) && defined(F_SETOWN)
	r = fcntl(fd, F_SETOWN, pid);
#elif defined(HAVE_IOCTL) && defined(FIOSETOWN)
	r = ioctl(fd, FIOSETOWN, &pid);
#else
	r = 0;
#endif
	if (r < 0) {
		return (r);
	}

#if defined(HAVE_FCNTL) && defined(O_ASYNC)
	r = fcntl(fd, F_GETFL, 0);
	r = fcntl(fd, F_SETFL, r|O_ASYNC);
#elif defined(HAVE_IOCTL) && defined(FIOASYNC)
	r = ioctl(fd, FIOASYNC, &on);
#else
	r = 0;
#endif
	if (r < 0) {
		return (r);
	}
#endif
	return (fd);
}

/*
 * Threaded create socket.
 */
int
threadedSocket(int af, int type, int proto)
{
	int fd;

	fd = socket(af, type, proto);
	return (threadedFileDescriptor(fd));
}

/*
 * Threaded file open.
 */
int
threadedOpen(const char* path, int flags, int mode)
{
	int fd;

	fd = open(path, flags, mode);
	return (threadedFileDescriptor(fd));
}

/*
 * Threaded socket connect.
 */
int
threadedConnect(int fd, struct sockaddr* addr, size_t len)
{
	int r;

#if !defined(BLOCKING_CALLS)
	r = connect(fd, addr, len);
	if ((r < 0) && (errno == EINPROGRESS || errno == EALREADY || errno == EWOULDBLOCK || errno == EINTR)) {
		blockOnFile(fd, TH_CONNECT);
		r = 0; /* Assume it's okay when we get released */
	}
#else
	do {
		r = connect(fd, addr, len);
	} while ((r < 0) && (errno == EINTR || errno == EALREADY));
	if (r < 0 && errno == EISCONN) {
		r = 0;
	}
#endif

	return (r);
}

/*
 * Threaded socket accept.
 */
int
threadedAccept(int fd, struct sockaddr* addr, size_t* len)
{
	int r;

	for (;;) {
#if defined(BLOCKING_CALLS)
		blockOnFile(fd, TH_ACCEPT);
#endif
		r = accept(fd, addr, len);
		if (r >= 0 || !(errno == EINPROGRESS || errno == EALREADY || errno == EWOULDBLOCK)) {
			break;
		}
#if !defined(BLOCKING_CALLS)
		blockOnFile(fd, TH_ACCEPT);
#endif
	}
	return (threadedFileDescriptor(r));
}

/*
 * Read but only if we can.
 */
ssize_t
threadedRead(int fd, void* buf, size_t len)
{
	ssize_t r;

#if defined(BLOCKING_CALLS)
	if (blockOnFile(fd, TH_READ) < 0) {
		return (-1);
	}
#endif
	for (;;) {
		r = read(fd, buf, len);
		if (r >= 0 || !(errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)) {
			return (r);
		}
		if (blockOnFile(fd, TH_READ) < 0) {
			return (-1);
		}
	}
}

/*
 * Write but only if we can.
 */
ssize_t
threadedWrite(int fd, const void * buf, size_t len)
{
	ssize_t r;
	const void* ptr;

	ptr = buf;
	r = 1;

	while (len > 0 && r > 0) {
#if defined(BLOCKING_CALLS)
		if (blockOnFile(fd, TH_WRITE) < 0) {
			return (-1);
		}
#endif
		r = write(fd, ptr, len);
		if (r >= 0) {
			ptr += r;
			len -= r;
		}
		else if (!(errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)) {
			return (-1);
		}
		else {
#if !defined(BLOCKING_CALLS)
			if (blockOnFile(fd, TH_WRITE) < 0) {
				return (-1);
			}
#endif
			r = 1;
		}
	}
	return (ptr - buf);
}
#endif
